(*** Simulation einer Evolution von ELEKs (Endliche Lebende Kluempchen). ***)
(*** Literatur:                                                          ***)
(*** [1] : "SPEKTRUM" 1/86, (A.K.Dewdney);                               ***)
(*** [2] : "Artificial Intelligence through Simulated Evolution"         ***)
(***       (Fogel, Owens, Walsh);                                        ***)
(***       Verlag:  John Wiley & Sons (New York,London,Sidney)           ***)
(*** [3] : John H. Holland, University of Michigan, ('60er Jahre)        ***)

PROGRAM elementary_soup (input,output);

  CONST max_period = 12;
        max_hit    = 100;

  TYPE  chromosome = ARRAY [1..10,0..15] OF integer;
        life_stat  = ARRAY [1..10]       OF (primitive,mutation,newcomer);

  VAR   best         : integer;            (* globales, bestes Elek-Ergebnis *)
        outfile      : text;       (* Festhalten der Origenes und des Besten *)
        chrom        : chromosome;                     (* Die 10 Chromosomen *)
        chrom_stat   : life_stat;                       (* Entwicklungsstufe *)
        status       : ARRAY [1..10]         OF integer;          (* Zustand *)
        score        : ARRAY [1..10]         OF integer;   (* Vorhersageerg. *)
        environment  : ARRAY [1..max_period] OF integer;       (* Umweltbed. *)

(*** Zufallszahlengenerator,  Algorithmus aus der  'HP-65 User Library' ***)
(*** random (0) liefert Wert 0 <= x < 1, random (z) liefert 1 <= x <= z ***)

FUNCTION random (range : integer) : real;
  VAR random_number : real;
BEGIN
  random_number := frac (9821 * random_number + 0.211327);
  IF range <> 0 THEN
    random := trunc (random_number * range) + 1
  ELSE
    random := random_number
END; (*** of random ***)

PROCEDURE increase (VAR argument : integer; step : integer);
BEGIN
  IF (step <> 1) THEN
    argument := argument + step
  ELSE
    argument := succ (argument)
END;  (*** of increase ***)

PROCEDURE swap (VAR arg1,arg2 : integer);
  VAR dummy : integer;
BEGIN
  dummy := arg1;
  arg1  := arg2;
  arg2  := dummy
END;  (*** of swap ***)

(*** Eingabe der Umweltbedingungen als Folge von '0' und '1' ***)

PROCEDURE create_environment_conditions;

  VAR cond   : STRING [max_period];           (* Umweltbedingung als Periode *)
      index,                                               (* Schleifenindex *)
      err    : integer;                        (* Fehlercode fuer Proc 'val' *)

BEGIN
  clrscr;
  writeln ('THE EVOLUTION OF ELEKS - A SIMULATION');
  writeln;
  REPEAT
    write ('Give me the Environment Condition (length: ',max_period:2,'): ');
    readln (cond);
    IF (length (cond) <> max_period) THEN
      writeln ('+++ illegal period length')
  UNTIL (length (cond) = max_period);
  (*** Aufschluesseln des Eingabestrings ***)
    FOR index := 1 TO max_period DO
      val (copy (cond,index,1),environment [index] ,err);
  clrscr;
  (*** Festhalten der Eingabe in einer Datei ***)
    writeln (outfile,'Environment Periode : ',cond);
    writeln (outfile)
END;  (*** of create_environment_conditions; ***)

(*** Initialisierung der 10 ELEKs durch zufaellige Werte ***)

PROCEDURE init_eleks;

  VAR idx1,                                (* Index fuer die einzelnen ELEKs *)
      idx2  : integer;       (* Index fuer den 'locus' des Gens im Chromosom *)

BEGIN
  FOR idx1 := 1 TO 10 DO
  BEGIN
    (*** File fuer naechstes ELEK vorbereiten ***)
      write (outfile,idx1:2,'. Elek : ');
    FOR idx2 := 0 TO 15 DO
    BEGIN
      IF odd (idx2) THEN
      BEGIN
        chrom [idx1,idx2] := trunc (random (4)) - 1;  (** 0(A) <= X <= 3(D) **)
        (*** Gen #idx2 (Zustand) d. Chromosoms von ELEK #idx1 wegschreiben ***)
          CASE chrom [idx1,idx2] OF
            0 : write (outfile,'A');
            1 : write (outfile,'B');
            2 : write (outfile,'C');
            3 : write (outfile,'D')
          END
      END
      ELSE
      BEGIN
        chrom [idx1,idx2] := trunc (random (2)) - 1;
        (*** Gen #idx2 (Ausgabe) d. Chromosoms von ELEK #idx1 wegschreiben ***)
          write (outfile,chrom [idx1,idx2])
      END
    END;
    writeln (outfile);
    (*** Initialisieren von ZUSTAND (status) und PUNKTE (score) ***)
      status [idx1] := 0;
      score  [idx1] := 0;
      chrom_stat [idx1] := primitive
  END
END;  (*** of init_eleks ***)

(*** Simulation der Evolution mit den 4 Modulen des Zyklus ***)

PROCEDURE evolution;

  VAR end_of_times : boolean;   (* 1 ELEK erreicht Vorhersagequote (max_hit) *)
      best_hit,                                           (* Bestes Ergebnis *)
      min_idx,                              (* Index des schlechtesten ELEKs *)
      max_idx,                                     (* Index des besten ELEKs *)
      chronons     : integer;                                      (* Zyklen *)

  PROCEDURE test_eleks;

    CONST no_of_env_sym = 100;    (* Entspricht einer Zeiteinheit (1 Zyklus) *)

    VAR   expected,                                      (* Ausgabe des Elek *)
          work_index,                       (* laufender Index: 1..no_of_sym *)
          elek_index,                              (* laufender Index: 1..10 *)
          symb,                     (* Ankommendes Eingabe- (Umwelt-) symbol *)
          next_symb,                  (* Aus Periode mach fortlaufende Reihe *)
          locus       : integer;       (* Kantenlaenge des Chromosomzugriffs *)

  BEGIN
    FOR work_index := 1 TO no_of_env_sym DO
    BEGIN
      (*** naechste Position fuer Umweltsymbol ***)
        next_symb := (work_index mod max_period) + 1;
      (*** aktuelles Umweltsymbol ***)
        symb := environment [next_symb];
      (*** naechstes (zu erwartendes) Umweltsymbol ***)
        IF (next_symb < 6) THEN
          expected := environment [next_symb + 1]
        ELSE
          expected := environment [1];
      FOR elek_index := 1 TO 10 DO
      BEGIN
        (*** Berechnen der Kantenlaenge fuer angesprochenes Gen ***)
          locus := 4 * status [elek_index] + 2 * symb;
        (*** Zustandsaenderung durchfuehren ***)
          status [elek_index] := chrom [elek_index,locus + 1];
        (*** Testen auf Vorhersage; falls richtig, inc (score) ***)
          IF (chrom [elek_index,locus] = expected) THEN
            increase (score [elek_index],1);
{       gotoxy (1,13 + elek_index);
        write (elek_index:2,' : ',status [elek_index]:2);
        write ('   ',score [elek_index]:2);
        CASE chrom_stat [elek_index] OF
          primitive : writeln (',  primitive');
          mutation  : writeln (',  mutation ');
          newcomer  : writeln (',  newcomer ')
        END}
      END
    END
  END;  (*** of test_eleks ***)

  (*** Berechnen des besten und des schlechtesten ELEKs ***)

  PROCEDURE extreme_eleks;

    VAR idx,                                                 (* Arbeitsindex *)
        min,                                       (* schlechtestes Ergebnis *)
        max  : integer;                                   (* bestes Ergebnis *)

  BEGIN
    max := score [1];
    min := score [1];
    min_idx := 1;
    max_idx := 1;
    FOR idx := 2 TO 10 DO
    BEGIN
      (*** Minimum ***)
        IF (score [idx] < min) THEN
        BEGIN
          min := score [idx];
          min_idx := idx
        END;
      (*** Maximum ***)
        IF (score [idx] > max) THEN
        BEGIN
          max := score [idx];
          max_idx := idx
        END
    END;
    best_hit := max;
    (*** Globales Sichern des besten ELEKs ueber seinen Index ***)
      best := max_idx;
    gotoxy (1,5);
    writeln ('Best  Elek #',max_idx:2,': ',max:4);
    writeln ('Worst Elek #',min_idx:2,': ',min:4)
  END;  (*** of extreme_eleks ***)

  (*** Crossing-Over des Besten (mother) und eines Zufaelligen (father) ***)

  PROCEDURE marriage_of_the_eleks;

    CONST love = 30;                  (* < 30: Paarung, >= 30: keine Paarung *)

    VAR   age,                                 (* Entscheidung ueber Paarung *)
          father,                                 (* Indexdes 1. Elternteils *)
          mother,                                (* Index des 2. Elternteils *)
          l,                                           (* locus im Chromosom *)
          c1,                            (* 1. Rekombinationsstelle (Anfang) *)
          c2      : integer;            (* 2. Rekombinationsstelle (Schluss) *)

  BEGIN
    age := trunc (random (69));
    IF (age > love) THEN
    BEGIN
      writeln ('     no wedding');
      writeln ('                ');
      {exit}
    END
    ELSE
    BEGIN
      writeln ('               ');
    (*** Auswahl der Eltern ***)
      mother := max_idx;
      REPEAT
        father := trunc (random (10))
      UNTIL (father <> mother);
    (*** Bestimmung des Umfangs der Rekombination ***)
      c1 := trunc (random (15));
      c2 := trunc (random (15));
    IF (c1 > c2) THEN
      swap (c1,c2);
    writeln ('c1,c2 : ',c1:4,c2:4);
    (*** Geburt eines neuen ELEKs und Tod des Schlechtesten ***)
      FOR l := 0 TO (c1 - 1) DO
        chrom [min_idx,l] := chrom [mother,l];
      FOR l := c1 TO (c2 - 1) DO
        chrom [min_idx,l] := chrom [father,l];
      FOR l := c2 TO 15 DO
        chrom [min_idx,l] := chrom [mother,l];
      chrom_stat [min_idx] := newcomer
    END
  END;  (*** of marriage_of_the_eleks ***)

  (*** Mutation infolge eines Treffers auf 1 Gen durch kosmisches Teilchen ***)

  PROCEDURE cosmic_nuke;

    VAR n,       (* Arbeitsindex fuer ELEK-Punkte (score) Neuinitialisierung *)
        target_elek,                                     (* getroffenes ELEK *)
        target_locus : integer;     (* getroffenes Gen im Chromosom des ELEK *)
        even_parity  : boolean;           (* Entscheidung ueber Art des Gens *)

  BEGIN
    target_elek  := trunc (random (10));
    target_locus := trunc (random (16) - 1);
    (*** Entscheidung: Zustands-Gen (ungerade) oder Ausgabe (gerade) ***)
      even_parity := NOT (odd (target_locus));
    (*** Zuteilung des neuen Gen-Wertes ***)
      IF even_parity THEN
        chrom [target_elek,target_locus] :=
                                  (chrom [target_elek,target_locus] + 1) mod 2
      ELSE
        chrom [target_elek,target_locus] :=
                                  (chrom [target_elek,target_locus] + 1) mod 4;
    writeln ('Hit on: Elek #',target_elek:2,', locus #',target_locus:2);
    IF chrom_stat [target_elek] <> newcomer THEN
      chrom_stat [target_elek] := mutation;
    (*** Neu-Initialisierung von score fuer naechste Generation ***)
      FOR n := 1 TO 10 DO
        score [n] := 0
  END;  (*** of cosmic_nuke ***)

  (*** Festhalten des Chromosoms des besten ELEKs in einer Datei ***)

  PROCEDURE best_elek;

    VAR i : integer;                 (* Arbeitsindex fuer locus im Chromosom *)

  BEGIN
    writeln (outfile);
    write (outfile,'Best Elek is : ');
    FOR i := 0 TO 15 DO
    BEGIN
      IF odd (i) THEN
      BEGIN
        CASE chrom [best,i] OF
          0 : write (outfile,'A');
          1 : write (outfile,'B');
          2 : write (outfile,'C');
          3 : write (outfile,'D')
        END
      END
      ELSE
      BEGIN
        write (outfile,chrom [best,i])
      END
    END;
    writeln (outfile);
    writeln (outfile,'Chronons: ',chronons)
  END;  (*** of best_elek ***)

  PROCEDURE show_chronons;
  BEGIN
    writeln ('Chronons : ',chronons:4)
  END;  (*** of show_chronons ***)

BEGIN  (*** of evolution ***)
  end_of_times := false;
  chronons := 0;
  WHILE NOT end_of_times DO
  BEGIN
    increase (chronons,1);
    test_eleks;
    extreme_eleks;
    marriage_of_the_eleks;
    cosmic_nuke;
    show_chronons;
    end_of_times := (best_hit >= max_hit)
  END;
  best_elek
END;  (*** of evolution ***)

(*** $MAIN: remote_soup ***)

BEGIN
  assign (outfile,'chromo.dat');
  rewrite (outfile);
  init_eleks;
  create_environment_conditions;
  evolution;
  close (outfile)
END.
